import hashlib
from typing import AnyStr


def md5(s: AnyStr, encoding="utf-8"):
    """
    return md5 hex string

    :param s: a string or bytes
    :param encoding: only used to encode s while s is a string
    """
    if isinstance(s, str):
        s = s.encode(encoding)
    return hashlib.md5(s).hexdigest()


def sha1(s: AnyStr, encoding="utf-8"):
    """
    return sha1 hex string

    :param s: a string or bytes
    :param encoding: only used to encode s while s is a string
    """
    if isinstance(s, str):
        s = s.encode(encoding)
    return hashlib.sha1(s).hexdigest()


def sha224(s: AnyStr, encoding="utf-8"):
    """
    return sha224 hex string

    :param s: a string or bytes
    :param encoding: only used to encode s while s is a string
    """
    if isinstance(s, str):
        s = s.encode(encoding)
    return hashlib.sha224(s).hexdigest()


def sha256(s: AnyStr, encoding="utf-8"):
    """
    return sha256 hex string

    :param s: a string or bytes
    :param encoding: only used to encode s while s is a string
    """
    if isinstance(s, str):
        s = s.encode(encoding)
    return hashlib.sha256(s).hexdigest()


def sha384(s: AnyStr, encoding="utf-8"):
    """
    return sha384 hex string

    :param s: a string or bytes
    :param encoding: only used to encode s while s is a string
    """
    if isinstance(s, str):
        s = s.encode(encoding)
    return hashlib.sha384(s).hexdigest()


def sha512(s: AnyStr, encoding="utf-8"):
    """
    return sha512 hex string

    :param s: a string or bytes
    :param encoding: only used to encode s while s is a string
    """
    if isinstance(s, str):
        s = s.encode(encoding)
    return hashlib.sha512(s).hexdigest()



import math
import random
import fractions
import _kzhutil
from .baillie_psw import baillie_psw


def normalize(array, **kwargs):
    """
    normalize the array to N(0, 1) by mean and std
    """
    mu, std = array.mean(**kwargs), array.std(**kwargs)
    return (array - mu) / std


def uniform_normalize(array, **kwargs):
    """
    normalize the array to U(0, 1)
    """
    ma, mi = array.max(**kwargs), array.min(**kwargs)
    return (array - mi) / (ma - mi)


primes = _kzhutil.primes


def exgcd(a, b):
    """
    get solution x, y for ax+by=gcd(a,b)

    :return: x, y, gcd(a, b)
    """
    def _exgcd(_a, _b):
        if _b == 0:
            return 1, 0, _a
        else:
            x, y, gcd = _exgcd(_b, _a % _b)
            x, y = y, (x - (_a // _b) * y)
            return x, y, gcd
    return _exgcd(a, b)


def inverse(x, n):
    """
    get the inverse element x^{-1} of x in Z_n where x^{-1} * x = 1 (mod n).
    gcd(x, n) must equal to 1 for the correct inverse, else return 0.

    :return: x^{-1}
    """
    inv, _, gcd = exgcd(x, n)
    return (inv + n) % n if gcd == 1 else 0


def continued_fraction(d, eps=1e-12, max_denominator=100000000):
    """
    calculate continued fraction coefficients [a0, a1, a2, ...] of d = a0 + 1/(a1 + 1/(a2 + 1/(a3 + ...)))
    """
    res, d = [int(d//1)], d % 1
    p_, q_, p, q = 1, 0, res[0], 1      # p/q is the convergent of continued fraction
    while True:
        a = int(1/d)
        p_, q_, p, q = p, q, a*p + p_, a*q + q_
        if q*(d*q + q_) > 1/eps or q > max_denominator: break     # error evaluate
        res.append(a)
        d = 1 / d % 1
    return res


def convergent_fraction(a):
    """
    get the convergent fraction of a cotinued fraction. e.g. [3, 7, 15, 1] -> 355/113
    """
    p_, q_, p, q = 1, 0, a[0], 1  # p/q is the convergent of continued fraction
    for ai in a[1:]:
        p_, q_, p, q = p, q, ai*p + p_, ai*q + q_
    return fractions.Fraction(p, q)


primes_1000 = set(primes(1000))
DEFAULT_MILLER_RABIN_K = 16


def miller_rabin(n, k=DEFAULT_MILLER_RABIN_K):
    """
    Miller-Rabin prime test
    :param n: the number
    :param k: how many miller-rabin test to try. the error rate is approximate to (1/4)^k.
              if k is 0, use deterministic miller test(deterministic if generalized Riemann hypothesis proved).
    :return: if n is a prime
    """
    if n <= 1000:
        return n in primes_1000
    if any(n % p == 0 for p in primes_1000):
        return False
    u, t = n-1, 0
    while u % 2 == 0:   # divide n-1 = u * 2^t
        u, t = u//2, t+1
    test_base = [random.randint(2, min(2**31, n-2)) for _ in range(k)] if k else range(2, min(n-2, int(math.log(n)**2*2)))
    for a in test_base:
        v = pow(a, u, n)    # a^u mod n
        if v == 1: continue
        for _ in range(t):
            if v == n - 1: break
            v = v * v % n
        else:
            return False
    return True


def next_prime(n, deterministic=False):
    """
    the next prime of n (n>2)
    """
    n += 1 if n % 2 == 0 else 2
    while not miller_rabin(n, 0 if deterministic else DEFAULT_MILLER_RABIN_K):
        n += 2
    return n


def random_prime(n, deterministic=False):
    """
    get a random prime contains n digits
    """
    num = 0
    for _ in range(n):
        num = num*10 + random.randint(0, 9)
    return next_prime(num, deterministic)


def factorization(n, deterministic=False):
    """
    return factors and exponents of n. (may run for real long time)
    n = p_1^{k_1}*p_2^{k_2}*p_3^{k_3}*... -> {p_1: k_1, p_2: k_2, p_3: k_3, ...}
    """
    factors, p = {}, 2
    while n > 1:
        if n % p == 0:
            factors[p] = factors.get(p, 0) + 1
            n //= p
        else:
            p = n if miller_rabin(n, 0 if deterministic else DEFAULT_MILLER_RABIN_K) else next_prime(p, deterministic)
    return factors


def lucas_test(n):
    """
    A deterministic primality test (may run for real long time)
    The running speed determined by whether n-1 is well factorized.
    """
    if n <= 1000:
        return n in primes_1000
    factors = {}
    for a in range(2, n):
        if pow(a, n-1, n) != 1:
            return False
        factors = factors if factors else factorization(n-1, True)
        for q in factors.keys():
            if pow(a, (n-1)//q, n) == 1:
                break
        else:
            return True
    return False


baillie_psw = baillie_psw


def euler_phi(n):
    ans = n
    for i in range(2, int(math.sqrt(n)) + 1):
        if n % i == 0:
            ans = ans // i * (i - 1)
            while n % i == 0:
                n = n // i
    if n > 1:
        ans = ans // n * (n - 1)
    return ans


def linear_interp(arr, x):
    import numpy as np
    x0 = np.floor(x).astype(int).clip(0, arr.shape[0] - 1)
    x1 = np.floor(np.asarray(x) + 1).astype(int).clip(0, arr.shape[0] - 1)
    return arr[x0] * (x1 - x) + arr[x1] * (x - x0)


def linear_interp2d(arr, x, y):
    import numpy as np
    x0 = np.floor(x).astype(int).clip(0, arr.shape[0] - 1)
    x1 = np.floor(np.asarray(x) + 1).astype(int).clip(0, arr.shape[0] - 1)
    y0 = np.floor(y).astype(int).clip(0, arr.shape[1] - 1)
    y1 = np.floor(np.asarray(y) + 1).astype(int).clip(0, arr.shape[1] - 1)
    return arr[x0, y0] * (x1 - x) * (y1 - y) + arr[x0, y1] * (x1 - x) * (y - y0) + \
           arr[x1, y0] * (x - x0) * (y1 - y) + arr[x1, y1] * (x - x0) * (y - y0)


class Triangle:
    def __init__(self, a=None, b=None, c=None, A=None, B=None, C=None, mode='rad'):
        assert mode in ('rad', 'deg'), "angle mode must be 'rad' or 'deg'"
        self.a, self.b, self.c = a, b, c
        self.A, self.B, self.C = (A % math.pi, B % math.pi, C % math.pi) if mode == 'rad' \
            else ((A % 180)*math.pi/180, (B % 180)*math.pi/180, (C % 180)*math.pi/180)
        self.mode = mode

    def __repr__(self):
        if self.mode == 'rad':
            return f"Triangle(a={self.a}, b={self.b}, c={self.c}, A={self.A/math.pi}*pi, B={self.B/math.pi}*pi, C={self.C/math.pi}*pi, mode={self.mode})"
        else:
            return f"Triangle(a={self.a}, b={self.b}, c={self.c}, A={self.A/math.pi*180}, B={self.B/math.pi*180}, C={self.C/math.pi*180}, mode={self.mode})"
    __str__ = __repr__

    def __eq__(self, other):
        return self.a == other.a and self.b == other.a and self.c == other.c

    @staticmethod
    def sss(a, b, c, mode='rad'):
        A = math.acos((b*b + c*c - a*a) / (2*b*c))
        B = math.acos((a*a + c*c - b*b) / (2*a*c))
        C = math.acos((b*b + a*a - c*c) / (2*b*a))
        A, B, C = (A, B, C) if mode == 'rad' else (A/math.pi*180, B/math.pi*180, C/math.pi*180)
        return Triangle(a, b, c, A, B, C, mode)

    @staticmethod
    def ssa(a, b, C, mode='rad'):
        C = C if mode == 'rad' else C*math.pi/180
        c = a*a + b*b - 2*a*b*math.cos(C)
        A = math.acos((b*b + c*c - a*a) / (2*b*c))
        B = math.acos((a*a + c*c - b*b) / (2*a*c))
        return Triangle(a, b, c, A, B, C, mode)

    @staticmethod
    def saa(a, B, C, mode='rad'):
        B, C = (B, C) if mode == 'rad' else (B*math.pi/180, C*math.pi/180)
        A = math.pi - B - C
        b = math.sin(B) / math.sin(A) * a
        c = math.sin(C) / math.sin(A) * a
        return Triangle(a, b, c, A, B, C, mode)



from typing import Callable, Sequence


class MappingSpace:
    def __init__(self, arr: Sequence, key: Callable):
        self.arr = arr
        self.key = key

    def __len__(self):
        return len(self.arr)

    def __getitem__(self, i):
        return self.key(self.arr[i])



from typing import AnyStr


def get_clipboard():
    """
    get clipboard text, works on win32

    :return: clipboard text
    """
    import win32clipboard
    import win32con
    win32clipboard.OpenClipboard()
    s = win32clipboard.GetClipboardData(win32con.CF_TEXT)
    win32clipboard.CloseClipboard()
    return s


def set_clipboard(s: AnyStr):
    """
    set clipboard text, works on win32

    :param s: clipboard text
    """
    import win32clipboard
    win32clipboard.OpenClipboard()
    win32clipboard.EmptyClipboard()
    win32clipboard.SetClipboardText(s)
    win32clipboard.CloseClipboard()



import json
import csv
import sys
import math
from kzhutil.functools import try_until
from typing import List, Dict, AnyStr


maxInt = sys.maxsize
try_until(csv.field_size_limit, [maxInt//(2**i) for i in range(int(math.log2(maxInt)))])


def read_from_file(fn, binary=False, encoding=None):
    """
    get all data of a file

    :param fn: file name or path
    :param binary: if true open in binary mode and return bytes
    :param encoding:
    :return: all data of the file
    """
    with open(fn, "r" if not binary else "rb", encoding=encoding) as f:
        return f.read()


def write_over_file(fn, s: AnyStr, binary=False, encoding=None):
    """
    overwrite a file

    :param fn: file name or path
    :param s: content to write
    :param binary: if true open in binary mode and s should be bytes
    :param encoding:
    """
    with open(fn, "w" if not binary else "wb", encoding=encoding) as f:
        return f.write(s)


def append_to_file(fn, s: AnyStr, binary=False, encoding=None):
    """
    append content to the end of a file

    :param fn: file name or path
    :param s: content to write
    :param binary: if true open in binary mode and s should be bytes
    :param encoding:
    """
    with open(fn, "a" if not binary else "ab", encoding=encoding) as f:
        return f.write(s)


def write_jsonl(fn, data: List[object], encoding=None, **json_kwargs):
    """
    write data to a jsonl file, where a long list is split into lines of json

    :param fn: file name or path
    :param data: a list data
    :param encoding:
    """
    with open(fn, 'w', encoding=encoding) as f:
        return f.write("\n".join([json.dumps(dl, **json_kwargs) for dl in data]))


def read_jsonl(fn, encoding=None, **json_kwargs) -> List[object]:
    """
    read from a jsonl file, where a long list is split into lines of json

    :param fn: file name or path
    :return: a list data
    :param encoding:
    """
    with open(fn, 'r', encoding=encoding) as f:
        return [json.loads(dl, **json_kwargs) for dl in f.readlines()]


def read_csv(fn, encoding=None, **csv_kwargs):
    """
    read a csv(comma-separated values) file

    :param fn: file name or path
    :return: list containing all data in .csv file
    :param encoding:
    """
    with open(fn, 'r', newline='', encoding=encoding) as f:
        return [*csv.reader(f, **csv_kwargs)]


def write_csv(fn, data: List[object], encoding=None, **csv_kwargs):
    """
    write to a file

    :param fn: file name or path
    :param data: data needed to be written to a csv file
    :param encoding:
    """
    with open(fn, 'w', newline='', encoding=encoding) as f:
        writer = csv.writer(f, **csv_kwargs)
        for row in data:
            writer.writerow(row)


def read_csv_dict(fn, encoding=None, **csv_kwargs):
    """
    read a csv(comma-separated values) file

    :param fn: file name or path
    :return: list containing all data in .csv file
    :param encoding:
    """
    with open(fn, 'r', newline='', encoding=encoding) as f:
        return [*csv.DictReader(f, **csv_kwargs)]


def write_csv_dict(fn, fieldnames, data: List[Dict], encoding=None, **csv_kwargs):
    """
    write to a file

    :param fn: file name or path
    :param fieldnames: fieldnames of the csv header
    :param data: data needed to be written to a csv file
    :param encoding:
    """
    with open(fn, 'w', newline='', encoding=encoding) as f:
        writer = csv.DictWriter(f, fieldnames=fieldnames, **csv_kwargs)
        writer.writeheader()
        for row in data:
            writer.writerow(row)



from typing import Iterable, Dict


def flatten(arr: Iterable):
    """
    [a, b, [c, [d, e], [f], g] -> [a, b, c, d, e, f, g]
    """
    return [i for item in arr for i in (flatten(item) if isinstance(item, Iterable) else [item])]


def flatten_dict(d: Dict, join=None) -> Dict:
    """
    {'a': 1, 'b': {'c': {'d': 2, 'e': 3}, 4: 5}, 'f': 6}
     -> {('a',): 1, ('b', 'c', 'd'): 2, ('b', 'c', 'e'): 3, ('b', 4): 5, 'f': 6}
    """
    res = {}
    for k, v in d.items():
        if isinstance(v, Dict):
            for ki, vi in flatten_dict(v, join).items():
                res.__setitem__((k,)+ki if join is None else str(k)+join+ki, vi)
        else:
            res.__setitem__((k,) if join is None else str(k), v)
    return res



import functools
import time
from typing import Callable, Iterable


def try_for(n):
    """
    try a function for n times till success
    """
    def decorator(f):
        @functools.wraps(f)
        def wrapped_f(*args, **kwargs):
            return try_for_(n, f, *args, **kwargs)
        return wrapped_f
    return decorator


def try_for_(n, f: Callable, *args, **kwargs):
    """
    try call f(*args, **kwargs) for n times till success
    """
    exception = None
    for _ in range(n):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            exception = e
    raise exception


def try_until(f: Callable, args: Iterable):
    """
    try f(*args[0]), f(*args[1]), f(*args[2]), ... until success then return
    """
    exception = None
    for arg in args:
        try:
            return f(*arg) if isinstance(arg, Iterable) else f(arg)
        except Exception as e:
            exception = e
    raise exception


def repeat_for(n):
    """
    repeat a function for n times
    """
    def decorator(f):
        @functools.wraps(f)
        def wrapped_f(*args, **kwargs):
            return repeat_for_(n, f, *args, **kwargs)
        return wrapped_f
    return decorator


def repeat_for_(n, f: Callable, *args, **kwargs):
    return [f(*args, **kwargs) for _ in range(n)]


def benchmark(n, f: Callable, *args, **kwargs):
    """
    repeat f(*args, **kwargs) n times and get the time
    """
    t0 = time.time()
    for _ in range(n):
        f(*args, **kwargs)
    return time.time()-t0



"""
author: armchaircaver
https://github.com/armchaircaver/Baillie-PSW/blob/main/baillie_psw.py

Implementation of the Baillie-PSW primality checking algorithm, from the following papers:

Lucas Pseudoprimes
Robert Baillie and Samuel S. Wagstaff, Jr.
Mathematics of Computation Vol. 35, No. 152 (Oct., 1980), pp. 1391-1417
https://www.jstor.org/stable/2006406

Strengthening the Baillie-PSW primality test,
Robert Baillie, Andrew Fiori and Samuel S. Wagstaff, Jr.
Math. Comp. 90 (2021), 1931-1955
https://arxiv.org/pdf/2006.14425v1.pdf,  https://homes.cerias.purdue.edu/~ssw/bfw.pdf

"""
import math


# from https://stackoverflow.com/questions/44531084/python-check-if-a-number-is-a-square
def is_square(n):
    if n <= 0:
        return n == 0
    x, y = 2**int(math.log2(n)//2), 2**int(math.log2(n)//2+1)
    while x + 1 < y:
        mid = (x + y) // 2
        if mid * mid < n:
            x = mid
        else:
            y = mid
    return n == x ** 2 or n == (x + 1) ** 2


# -------------------------------------------------------------------------------

# adapted from https://rosettacode.org/wiki/Jacobi_symbol#Python
def jacobi(a, n):
    a %= n
    result = 1
    while a != 0:
        while a % 2 == 0:
            a //= 2
            n_mod_8 = n % 8
            if n_mod_8 in (3, 5):
                result = -result
        a, n = n, a
        if a % 4 == 3 and n % 4 == 3:
            result = -result
        a %= n
    if n == 1:
        return result
    else:
        return 0


# -------------------------------------------------------------------------------
def miller_rabin_base_2(n):
    # adapted from https://rosettacode.org/wiki/Miller%E2%80%93Rabin_primality_test#Python
    s = 0
    d = n - 1
    while d % 2 == 0:
        d >>= 1
        s += 1

    def trial_composite(a):
        if pow(a, d, n) == 1:
            return False
        for i in range(s):
            if pow(a, 2 ** i * d, n) == n - 1:
                return False
        return True

    return not trial_composite(2)


# -------------------------------------------------------------------------------

def D_chooser(n):
    # Choose a D value suitable for the Baillie-PSW test
    D = 5
    j = jacobi(D, n)

    while j > 0:
        D += 2 if D > 0 else -2
        D *= -1

        if D == -15:
            # check for a square
            if is_square(n):
                # The value of D isn't 0, but we are just communicating
                # that we have found a square
                return (0, 0)

        j = jacobi(D, n)
    return (D, j)


# -------------------------------------------------------------------------------

def div2mod(x, n):
    # divide by 2 modulo n
    # assumes n is odd
    if x & 1:
        return ((x + n) >> 1) % n
    return (x >> 1) % n


# -------------------------------------------------------------------------------

def U_V_subscript(k, n, P, D):
    U = 1
    V = P
    digits = bin(k)[2:]

    for digit in digits[1:]:
        U, V = (U * V) % n, div2mod(V * V + D * U * U, n)

        if digit == '1':
            U, V = div2mod(P * U + V, n), div2mod(D * U + P * V, n)
    return U, V


# -------------------------------------------------------------------------------

def lucas_pp(n, D, P, Q):
    assert n & 1
    U, V = U_V_subscript(n + 1, n, P, D)
    return U == 0


# -------------------------------------------------------------------------------

def lucas_spp(n, D, P, Q):
    # Lucas strong probable prime test
    # https://arxiv.org/pdf/2006.14425v1.pdf
    # This is a bit slower than lucas_pp, so is not used

    assert n & 1

    d = n + 1
    s = 0
    while (d & 1) == 0:
        s += 1
        d >>= 1

    U, V = U_V_subscript(d, n, P, D)
    if U == 0:
        return True

    Q = pow(Q, d, n)

    for r in range(s):
        if V == 0:
            return True
        V = (V * V - 2 * Q) % n
        Q = pow(Q, 2, n)

    return False


# -------------------------------------------------------------------------------
primes_100 = {2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97}


def baillie_psw(n):
    if n <= 100:
        return n in primes_100
    if any(n % p == 0 for p in primes_100):
        return False

    if not miller_rabin_base_2(n):
        return False

    D, j = D_chooser(n)

    if j == 0:
        return False  # see https://www.jstor.org/stable/2006406

    # even numbers and squares have been eliminated by this point
    return lucas_pp(n, D, 1, (1 - D) // 4)  # slightly faster than lucas_spp



from setuptools import setup
from pybind11.setup_helpers import Pybind11Extension


ext_modules = [
    Pybind11Extension(
        "_kzhutil",
        ['src/math.cpp'],  # Sort source files for reproducibility
    ),
]


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='kzhutil',
    version='0.0.4',
    author='kuangzh',
    author_email='kuangzh@smail.nju.edu.cn',
    url='https://github.com/kuangkzh/kzhutil',
    description='some python utils',
    long_description=long_description,
    long_description_content_type="text/markdown",
    ext_modules=ext_modules,
    zip_safe=False,
)



import shutil


def load_pretrained(model_class, model_name, model_path, **kwargs):
    """
    load pretrained transformers. If a saved model provided, then load directly. Else load from hub and save.

    :param model_class: e.g. transformers.AutoModel, transformers.AutoTokenizer
    :param model_name: the transformer model name
    :param model_path: try to load a model in this path
    """
    cache_path = model_path + "/cache"
    try:
        model = model_class.from_pretrained(model_path, cache_dir=cache_path, **kwargs)
    except Exception:
        shutil.rmtree(cache_path)
        model = model_class.from_pretrained(model_name, cache_dir=cache_path, **kwargs)
        model.save_pretrained(model_path)
        shutil.rmtree(cache_path)
    return model
